import json
import re
import sys

import click
from rich.console import Console
from rich.table import Table

from delta.cli.utils import API, Utils, ReturnCode

DEFAULT_LIMIT = 15


@click.help_option("--help", "-h")
@click.command(
    name='get',
    short_help='Retrieve an artifact')
@click.option(
    '--conf',
    '-c',
    type=str,
    default=None,
    help='Path to the conf file')
@click.option(
    '--download',
    '-d',
    type=bool,
    required=False,
    is_flag=True,
    default=False)
@click.option(
    '--file',
    '-F',
    type=str,
    default=None,
    help='Path of the file to save the artifact, by default '
         'it is the basename of the artifact in current directory')
@click.option(
    '--format-output',
    '-f',
    type=str,
    default=None,
    help='Format of the output json/text default is text')
@click.argument('artifact_id')
def get_artifact(conf, artifact_id, download,
                 file: str, format_output) -> None:
    """Retrieve an artifact and store it in a file.

    ARTIFACT_ID: Id of the Artifact [MANDATORY]

    Example:

    deltatwin drive artifact get b8810ff1-16c7-4269-b784-0c5ce392ff25
    --file my_art.jpg --download

    deltatwin drive artifact get b8810ff1-16c7-4269-b784-0c5ce392ff25
    """

    if download:
        resp = API.download_artifact(conf, artifact_id)

        if file is None:
            d = resp.headers['content-disposition']
            file = re.findall("filename=\"(.+)\"", d)[0]

        with open(file, 'wb') as f:
            for chunk in resp.iter_content(chunk_size=8192):
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                # if chunk:
                f.write(chunk)

        click.echo(f"{Utils.log_info} Artifact \"{artifact_id}\" successfully "
                   f"downloaded in \"{file}\"")
    else:
        resp = API.get_artifact(conf, artifact_id)

        data = {
            'artefact_id': resp['artefact_id'],
            'name': resp['name'],
            'twin_name': resp['twin_name'],
            'content_type': resp['content_type'],
            'content': resp['content'],
            'size': resp['size'],
            'description': resp['description'],
            'topics': resp['topics']
        }

        if format_output is not None and format_output.lower() == 'json':
            click.echo(json.dumps(data, indent=4))
        else:
            table = Table()
            table.add_column("Artifact ID", no_wrap=True)
            table.add_column("Artifact Name")
            table.add_column("DeltaTwin Name")
            table.add_column("Content")
            table.add_column("Content type")
            table.add_column("Size")
            table.add_column("Description")
            table.add_column("Topics")

            table.add_row(str(data['artefact_id']),
                          data['name'],
                          data['twin_name'],
                          str(data['content']),
                          data['content_type'],
                          str(data['size']),
                          data['description'],
                          str(data['topics']))

            console = Console()
            console.print(table)
